#! /usr/bin/env perl
###########################################################################
# $Id: ReEstimate,v 1.2 2000/04/13 12:16:27 ftj Exp $
#
# Usage: ReEstimate srchmm trghmm cfgfile scpfile alignmentfile labext modellist logfile
#
# Perform re-estimation with multiple simultaneous processes.
# The training set will be split into a minimum of 4 pieces, which will
# be processed in parallel. Large training sets will be split into more,
# so that the maximum piece size is 30k sentences. 
#
# Todo:
#
# * The number of parallel processes should depend on available CPU and RAM.
# * Sequential processing of pieces should be used to avoid HEREst problem.
#
# History:
#
# Unchanged in Masper - MonoTrain
#
# $Log: ReEstimate,v $
# Revision 1.2  2000/04/13 12:16:27  ftj
# Added proper synchronisation of subprocesses, using fork instead of
# background processes. This Fixed the "CheckMarker:" error message from HERest.
#
# Revision 1.1  2000/02/04 11:46:50  ftj
# Combined trntools and tsttools into a single tools directory
#
# Revision 1.2  1999/12/10 11:50:57  ftj
# split option bug fixed. Reported by Narada Warakagoda and Giampiero Salvi.
#
# Version 1.0  19.10.99  Borge Lindberg
#
###########################################################################

die "Wrong number of arguments (8 expected). Stopped" unless ($#ARGV == 7);
$sourcehmm = shift(@ARGV);
$targethmm = shift(@ARGV);
$cfgfile   = shift(@ARGV);
$scpfile   = shift(@ARGV);
$alignfile = shift(@ARGV);
$labext	   = shift(@ARGV);
$modellist = shift(@ARGV);
$logfile   = shift(@ARGV);

$maxpiecesize      = 30000; 	# The exact limit depends on utterance lengths
$MinimumParallel   = 4;		# Minimum number of parallel training processes

###########################################################################

$| = 1; # avoid multiple copies of log messages

# Split trainset into pieces

$piecesize = `wc -l < $scpfile`;
$nr_splits = int($piecesize/$maxpiecesize) + 1;
$nr_splits = $MinimumParallel if ($nr_splits < $MinimumParallel);
$piecesize = int($piecesize/$nr_splits) + 1;
&run ("rm -f ${scpfile}_piece_*");                         # delete previous piece files 
&run ("split -l $piecesize  $scpfile ${scpfile}_piece_");  # trainset splitting 


# Process trainset pieces (in parallel)

print " ...reestimating to $targethmm ($nr_splits parallel processes) ",`date`;
&run ("cp $sourcehmm $targethmm"); # initial model file copy
$piecenum = 0;
@pieces = `ls ${scpfile}_piece_*`;
&run ("rm -f $logfile.?"); # remove all logfile pieces
&run ("rm -f HER*.acc");   # remove all accumulator files
foreach $piece (@pieces) {
  chop($piece);
  $piecenum+=1;
  unless ($pid = fork) {   # child process: start reestimation
    die "Cannot fork: $!" unless defined $pid;
    &run("HERest -C $cfgfile -I $alignfile -X $labext ".
	  "-t 250.0 150.0 1000.0 -S $piece ".
	  "-p $piecenum -B -H $targethmm $modellist > $logfile.$piecenum");
    exit; # child only
  }
  $pids{$piece} = $pid;    # parent process: record child pid and continue forking
}


# Wait for children to complete

foreach $piece (@pieces) {
  $pid = $pids{$piece};
  waitpid($pid,0);
}


# Merge accumulators

&run ("cat $logfile.? > $logfile");
&run ("rm -f $logfile.?");
&run ("HERest -C $cfgfile -I $alignfile -X $labext ".
      "-t 250.0 150.0 1000.0 ".
      "-p 0 -H $targethmm $modellist HER*.acc >> $logfile");
&run ("rm -f HER*.acc");  
print " ...finished ",`date`;

###########################################################################

sub run { # run unix shell command with error check
  local($cmd) = @_;
  die "Unix command failed:\n $cmd\n" if system "$cmd";
}

